# -*- coding: utf-8 -*-
"""

"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

__author__ = "timmyliang"
__email__ = "820472580@qq.com"
__date__ = "2020-11-19 00:05:51"

import six
import inspect
from .binding import Binding
from .binder import Binder, BinderBase
from .util import ListGet
from Qt import QtCore, QtWidgets


class ItemMeta(type(QtCore.QObject)):
    def __new__(cls, name, bases, attrs):
        init = attrs.get("__init__")
        if init:
            binder_dict = {
                name: binder
                for name, binder in attrs.items()
                if isinstance(binder, BinderBase)
            }
            attrs["__init__"] = cls.init_deco(init, binder_dict)

        return super(ItemMeta, cls).__new__(cls, name, bases, attrs)

    @classmethod
    def init_deco(cls, func, binder_dict):
        @six.wraps(func)
        def wrapper(self, *args, **kwargs):
            # NOTE inject class binder to self binder
            # TODO FnBinding point to self
            cls = self.__class__
            name = ""
            for name, binder in binder_dict.items():
                _binder = Binder()
                for k, v in binder._var_dict_.items():
                    setattr(_binder, k, v)
                setattr(self, name, _binder)

            if not cls.__binder__:
                cls.__binder__ = name

            return func(self, *args, **kwargs)

        return wrapper


class ItemMixin(six.with_metaclass(ItemMeta)):
    __binder__ = ""
    __data__ = {}
    __items__ = []
    __filter__ = []
    __layout__ = None



# -*- coding: utf-8 -*-
"""
# TODO create a Vuex like pannel
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

__author__ = 'timmyliang'
__email__ = '820472580@qq.com'
__date__ = '2020-11-07 22:24:50'

import os
import re
import sys
import time
from textwrap import dedent

from Qt import QtGui
from Qt import QtCore
from Qt import QtWidgets

from .util import CollapsibleWidget

DIR = os.path.dirname(__file__)

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(393, 713)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setObjectName("verticalLayout")
        self.Var_Toggle = QtWidgets.QPushButton(Form)
        self.Var_Toggle.setObjectName("Var_Toggle")
        self.verticalLayout.addWidget(self.Var_Toggle)
        self.Filter_Table = FilterTableWidget(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.Filter_Table.sizePolicy().hasHeightForWidth())
        self.Filter_Table.setSizePolicy(sizePolicy)
        self.Filter_Table.setMinimumSize(QtCore.QSize(0, 0))
        self.Filter_Table.setObjectName("Filter_Table")
        self.verticalLayout.addWidget(self.Filter_Table)
        self.Scope_Toggle = QtWidgets.QPushButton(Form)
        self.Scope_Toggle.setObjectName("Scope_Toggle")
        self.verticalLayout.addWidget(self.Scope_Toggle)
        self.Scope_List = QtWidgets.QListWidget(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.Scope_List.sizePolicy().hasHeightForWidth())
        self.Scope_List.setSizePolicy(sizePolicy)
        self.Scope_List.setMinimumSize(QtCore.QSize(0, 0))
        self.Scope_List.setObjectName("Scope_List")
        self.verticalLayout.addWidget(self.Scope_List)
        spacerItem = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.verticalLayout.setStretch(0, 1)
        self.verticalLayout.setStretch(2, 1)

        self.retranslateUi()
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self):
        # Form.setWindowTitle(QtWidgets.QApplication.translate("Form", "Form", None, -1))
        self.Var_Toggle.setText(QtWidgets.QApplication.translate("Form", "Scope Variable", None, -1))
        self.Scope_Toggle.setText(QtWidgets.QApplication.translate("Form", "Function Scope", None, -1))
        
        
class LineEditDelegate(QtWidgets.QStyledItemDelegate):

    moveCurrentCellBy = QtCore.Signal(int, int)
    varaibleModified = QtCore.Signal(QtWidgets.QStyledItemDelegate,str)

    def __init__(self, parent=None):
        super(LineEditDelegate, self).__init__(parent)
        self.modified = True
        
    def createEditor(self, parent, option, index):
        self.editor = QtWidgets.QLineEdit(parent)
        self.editor.setFrame(False)
        self.editor.installEventFilter(self)
        self.editor.modelIndex = index
        return self.editor

    def setEditorData(self, editor, index):
        """setEditorData 
        Get cell data to lineEdit 
        
        :param editor: lineEdit widget
        :type editor: QLineEdit
        :param index: cell index
        :type index: QtCore.QModelIndex
        """
        value = index.model().data(index, QtCore.Qt.EditRole)
        editor.setText(value)

    def setModelData(self, editor, model, index):
        value = editor.text()
        model.setData(index, value, QtCore.Qt.EditRole)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)

    def eventFilter(self, target, event):
        if target is self.editor:
            # NOTE 点击其他地方 取消修改
            if event.type() == QtCore.QEvent.Type.FocusAboutToChange:
                self.closeEditor.emit(self.editor, QtWidgets.QAbstractItemDelegate.NoHint)
                
            if event.type() == QtCore.QEvent.KeyPress:

                if event.key() in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter):
                    # NOTE 变量修改
                    self.varaibleModified.emit(self,self.editor.text())
                    if self.modified:
                        self.commitData.emit(self.editor)
                    else:
                        self.modified = True

                    self.closeEditor.emit(self.editor, QtWidgets.QAbstractItemDelegate.NoHint)

                if self.editor.hasFocus():
                    # NOTE 阻断上下箭头的 Layout
                    if event.key() in [QtCore.Qt.Key_Down,QtCore.Qt.Key_Up]:
                        return True
                    return False

                moveCell, row, column = False, 0, 0
                if event.key() == QtCore.Qt.Key_Down:
                    moveCell, row, column = True, 1, 0
                if event.key() == QtCore.Qt.Key_Up:
                    moveCell, row, column = True, -1, 0
                if event.key() in (QtCore.Qt.Key_Right, QtCore.Qt.Key_Tab):
                    moveCell, row, column = True, 0, 1
                if event.key() in (QtCore.Qt.Key_Left, QtCore.Qt.Key_Backtab):
                    moveCell, row, column = True, 0, -1

                if moveCell:
                    self.commitData.emit(self.editor)
                    self.closeEditor.emit(self.editor, QtWidgets.QAbstractItemDelegate.NoHint)
                    self.moveCurrentCellBy.emit(row, column)
                    return True

        return False   

class FilterTableWidget(QtWidgets.QWidget):
    """
    # NOTE https://stackoverflow.com/questions/14068823/how-to-create-filters-for-qtableview-in-pyqt
    """
    def __init__(self, parent=None):
        super(FilterTableWidget, self).__init__(parent)
        self.lineEdit = QtWidgets.QLineEdit(self)
        self.view     = QtWidgets.QTableView(self)
        self.comboBox = QtWidgets.QComboBox(self)
        self.label    = QtWidgets.QLabel(self)

        # NOTE 延长Header
        self.view.horizontalHeader().setStretchLastSection(True)            

        self.view.setWordWrap(True)
        self.view.setTextElideMode(QtCore.Qt.ElideMiddle)
        self.view.resizeRowsToContents()

        self.delegate = LineEditDelegate()
        self.view.setItemDelegate(self.delegate)

        self.gridLayout = QtWidgets.QGridLayout(self)
        self.gridLayout.setContentsMargins(0,0,0,0)
        self.gridLayout.addWidget(self.lineEdit, 0, 1, 1, 1)
        self.gridLayout.addWidget(self.view, 1, 0, 1, 3)
        self.gridLayout.addWidget(self.comboBox, 0, 2, 1, 1)
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.comboBox.setMinimumWidth(150)
        self.model = QtGui.QStandardItemModel(self)

        self.proxy = QtCore.QSortFilterProxyModel(self)
        self.proxy.setSourceModel(self.model)

        self.view.setModel(self.proxy)

        self.lineEdit.textChanged.connect(self.on_lineEdit_textChanged)
        self.comboBox.currentIndexChanged.connect(self.on_comboBox_currentIndexChanged)

        self.horizontalHeader = self.view.horizontalHeader()
        self.horizontalHeader.sectionClicked.connect(self.on_view_horizontalHeader_sectionClicked)

        self.retranslateUi()
    
    def changeEvent(self, event):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslateUi()
        super(FilterTableWidget, self).changeEvent(event)

    def retranslateUi(self):
        self.label.setText(QtWidgets.QApplication.translate("label", "Regular Expression"))

        # NOTE 修改行名称
        header_list = [
            QtWidgets.QApplication.translate("header", "Variable Name"),
            QtWidgets.QApplication.translate("header", "Variable Value"),
        ]
        for column,label in enumerate(header_list):
            self.model.setHeaderData(column, QtCore.Qt.Horizontal,label)

        self.comboBox.clear()
        self.comboBox.addItems(header_list)
    
    def addItems(self,data):

        for var,val in data.items():
            var_item = QtGui.QStandardItem(str(var))
            var_item.setEditable(False)
            var_item.setToolTip(str(type(val)))


            val_item = QtGui.QStandardItem('"%s"' % val if type(val) == str else str(val))


            self.model.invisibleRootItem().appendRow([var_item,val_item])

        self.retranslateUi()

    def clearItems(self):
        for i in range(self.model.rowCount()):
            self.model.removeRows(i,self.model.rowCount())

    def setModel(self,model):
        self.model = model
        self.proxy.setSourceModel(self.model)

    @QtCore.Slot(int)
    def on_view_horizontalHeader_sectionClicked(self, logicalIndex):
        self.logicalIndex   = logicalIndex
        self.menuValues     = QtWidgets.QMenu(self)
        self.signalMapper   = QtCore.QSignalMapper(self)  

        self.comboBox.blockSignals(True)
        self.comboBox.setCurrentIndex(self.logicalIndex)
        self.comboBox.blockSignals(True)

        valuesUnique = [    self.model.item(row, self.logicalIndex).text()
                            for row in range(self.model.rowCount())
                            ]

        actionAll = QtWidgets.QAction("All", self)
        actionAll.triggered.connect(self.on_actionAll_triggered)
        self.menuValues.addAction(actionAll)
        self.menuValues.addSeparator()

        for actionNumber, actionName in enumerate(sorted(list(set(valuesUnique)))):              
            action = QtWidgets.QAction(actionName, self)
            self.signalMapper.setMapping(action, actionNumber)  
            action.triggered.connect(self.signalMapper.map)  
            self.menuValues.addAction(action)

        self.signalMapper.mapped.connect(self.on_signalMapper_mapped)  

        self.headerPos = self.view.mapToGlobal(self.horizontalHeader.pos())        

        posY = self.headerPos.y() + self.horizontalHeader.height()
        posX = self.headerPos.x() + self.horizontalHeader.sectionPosition(self.logicalIndex)

        self.menuValues.exec_(QtCore.QPoint(posX, posY))

    @QtCore.Slot()
    def on_actionAll_triggered(self):
        filterColumn = self.logicalIndex
        filterString = QtCore.QRegExp(  "",
                                        QtCore.Qt.CaseInsensitive,
                                        QtCore.QRegExp.RegExp
                                        )

        self.proxy.setFilterRegExp(filterString)
        self.proxy.setFilterKeyColumn(filterColumn)

    @QtCore.Slot(int)
    def on_signalMapper_mapped(self, i):
        stringAction = self.signalMapper.mapping(i).text()
        filterColumn = self.logicalIndex
        filterString = QtCore.QRegExp(  stringAction,
                                        QtCore.Qt.CaseSensitive,
                                        QtCore.QRegExp.FixedString
                                        )

        self.proxy.setFilterRegExp(filterString)
        self.proxy.setFilterKeyColumn(filterColumn)

    @QtCore.Slot(str)
    def on_lineEdit_textChanged(self, text):
        search = QtCore.QRegExp(    text,
                                    QtCore.Qt.CaseInsensitive,
                                    QtCore.QRegExp.RegExp
                                    )

        self.proxy.setFilterRegExp(search)

    @QtCore.Slot(int)
    def on_comboBox_currentIndexChanged(self, index):
        self.proxy.setFilterKeyColumn(index)



class InfoPanel(QtWidgets.QWidget,Ui_Form):
    def __init__(self):
        super(InfoPanel,self).__init__()

        self.setupUi(self)
        # ui_path = os.path.join(DIR,"ui","debugVar.ui")
        # loadUi(ui_path,self)

        self.var_toggle_anim = CollapsibleWidget.install(self.Var_Toggle,self.Filter_Table)
        self.scope_toggle_anim = CollapsibleWidget.install(self.Scope_Toggle,self.Scope_List)
    
    def initialize(self):
        self.Scope_List.itemClicked.connect(self.itemClickEvent)
        self.Filter_Table.delegate.varaibleModified.connect(self.modifyScopeEvent)

    def modifyScopeEvent(self,delegate,val):
        delegate.modified = True
        model_index = delegate.editor.modelIndex
        row = model_index.row()
        v_model = self.Filter_Table.view.verticalHeader().model()
        row_label = v_model.headerData(row, QtCore.Qt.Vertical) 

        model = self.Filter_Table.model
        var_name = model.item(row_label-1, 0).text()
        item = self.Scope_List.currentItem()

        globals = item.globals
        locals = item.locals
        # try:
        #     code = "%s = %s\n" % (var_name,val)
        #     code = compile(code, '<stdin>', 'single')
        #     exec code in globals, locals

        #     val = item.locals[var_name]
        #     val = '"%s"' % val if type(val) == str else str(val)
        #     delegate.editor.setText(val)
        # except:
        #     t, v = sys.exc_info()[:2]
        #     if type(t) == type(''):
        #         exc_type_name = t
        #     else: exc_type_name = t.__name__
        #     print >>sys.stdout, '***', exc_type_name + ':', v
        #     delegate.modified = False

    def retranslateUi(self):
        drop_icon = self.Var_Toggle.text()[:1]
        self.Var_Toggle.setText(drop_icon + QtWidgets.QApplication.translate("info","Scope Variable"))
        drop_icon = self.Scope_Toggle.text()[:1]
        self.Scope_Toggle.setText(drop_icon + QtWidgets.QApplication.translate("info","Function Scope"))

    def changeEvent(self, event):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslateUi()
        super(InfoPanel, self).changeEvent(event)

    def itemClickEvent(self,item):
        self.Filter_Table.clearItems()
        self.Filter_Table.addItems(item.locals)
        
        # NOTE 定位代码文件
        path = item.text()
        lineno = re.search("\d+$",path[:-1]).group(0)
        filename = path[:-len(lineno)-2]
        
        # # NOTE 代码显示
        # if not os.path.exists(filename):
        #     code = "%s %s" % (filename , QtWidgets.QApplication.translate("reading", "file not exists"))
        # elif self.panel.link.path == filename:
        #     code = ''
        # else:
        #     try:
        #         with open(filename,'r') as f:
        #             code = f.read()
        #     except:
        #         code = "%s %s" % (filename , QtWidgets.QApplication.translate("reading", "read fail"))
        
        # self.panel.link.setText(filename,lineno)
        # self.panel.editor.setPlainText(code) if code else None
        # self.panel.editor.paintLine(int(lineno))
        
        # self.Filter_Table.lineEdit.clear()

    def addItems(self,data):
        self.Filter_Table.addItems(data)

    def clear(self):
        self.Filter_Table.clearItems()

class PanelEvent(QtCore.QObject):
    
    def __init__(self):
        super(PanelEvent,self).__init__()
        event = QtCore.QEvent(QtCore.QEvent.User)
        QtWidgets.QApplication.postEvent(self, event)
        self.installEventFilter(self)

    def eventFilter(self,receiver,event):
        global __QBinder_info_panel__
        __QBinder_info_panel__ = InfoPanel()
        __QBinder_info_panel__.show()
        return False

def show_info_panel():
    # NOTE show pannel with event loop 
    global __QBinder_info_panel__
    __QBinder_info_panel__ = PanelEvent()

if __name__ == "__main__":
    
    app = QtWidgets.QApplication(sys.argv)

    widget = InfoPanel()
    widget.show()

    sys.exit(app.exec_())
    


# -*- coding: utf-8 -*-
"""

"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

__author__ = "timmyliang"
__email__ = "820472580@qq.com"
__date__ = "2020-11-04 15:30:25"

import os
import six
import sys
import time
import uuid
import json
import random
import hashlib
import inspect
import tempfile
from functools import partial
from collections import OrderedDict
from Qt import QtCore, QtWidgets
from .binding import Binding,FnBinding, BindingProxy
from .util import nestdict, defaultdict
from collections import OrderedDict
from .eventhook import QEventHook, Iterable

event_hook = QEventHook()


class BinderCollector(QtCore.QObject):
    Binders = OrderedDict()
    __flag__ = True

    def get_current_Binders(self):
        uid = uuid.uuid4()
        if self.Binders:
            last_uid = list(self.Binders.keys())[-1]
            if self.__flag__:
                uid = last_uid
                self >> event_hook(QtCore.QEvent.User, lambda: self.set_flag(False))
                event = QtCore.QEvent(QtCore.QEvent.User)
                QtWidgets.QApplication.postEvent(self, event)
            else:
                self.set_flag(True)

        BinderCollector.Binders.setdefault(uid, [])
        return BinderCollector.Binders[uid]

    @classmethod
    def set_flag(cls, flag):
        cls.__flag__ = flag

    @classmethod
    def get_last_key(cls):
        return list(cls.Binders.keys())[-1]

class BinderDumper(QtCore.QObject):
    _dumper_dict_ = {}
    __init_flag = False

    def __new__(cls, binder, *args, **kwargs):
        instance = cls._dumper_dict_.get(id(binder))
        cls.__init_flag = not instance
        if cls.__init_flag:
            instance = super(BinderDumper, cls).__new__(cls, binder, *args, **kwargs)
            cls._dumper_dict_[id(binder)] = instance
        
        return instance

    def __init__(self, binder, db_name, filters=None):
        if not self.__init_flag:
            return
        super(BinderDumper, self).__init__()
        self.binder = binder
        self.db_name = db_name
        self._filters_ = (
            {filters}
            if isinstance(filters, str)
            else set(filters)
            if isinstance(filters, Iterable)
            else set()
        )
        folder = os.path.join(tempfile.gettempdir(), "QBinder")
        if not os.path.isdir(folder):
            os.mkdir(folder)
        self.path = os.path.join(folder, "%s.json" % db_name)

        # NOTE using timer call the __prepare__ in the next evnet loop (for loading delay)
        self >> event_hook(
            QtCore.QEvent.User, lambda: QtCore.QTimer.singleShot(0, self.__prepare__)
        )
        event = QtCore.QEvent(QtCore.QEvent.User)
        QtWidgets.QApplication.postEvent(self, event)
        
        self.auto_load = True

    def set_auto_load(self,enabeld):
        self.auto_load = enabeld

    def __prepare__(self):
        if self.auto_load:
            for k, binding in self.binder._var_dict_.items():
                if k in self._filters_:
                    binding.connect(self.save)
            self.load()

    def __enter__(self):
        BinderBase._trace_flag_ = True
        return self

    def __exit__(self, *args):
        BinderBase._trace_flag_ = False
        self._filters_.update(BinderBase._trace_setattr_)

    def __del__(self):
        self.__dumper_list.remove(self)
        super(BinderDumper, self).__del__()

    def save(self, path="", indent=None):
        path = path if path else self.path
        data = {
            k: v.val for k, v in self.binder._var_dict_.items() if k in self._filters_
        }
        with open(path, "w") as f:
            json.dump(data, f, indent=indent)

    def load(self, path=""):
        path = path if path else self.path
        if not os.path.exists(path):
            return
        QtCore.QTimer.singleShot(0, partial(self.read, path))

    def read(self, path):
        try:
            with open(path, "r") as f:
                data = json.load(f, encoding="utf-8")
            for k, v in data.items():
                setattr(self.binder, k, v)
        except:
            # NOTE May be the file broken
            os.remove(path)


class BinderDispatcher(QtCore.QObject):
    _trace_dict_ = nestdict()
    __instance = None
    __init_flag = False

    def __new__(cls, binder):
        cls.binder = binder
        if cls.__instance is None:
            cls.__instance = super(BinderDispatcher, cls).__new__(cls)
            cls.__init_flag = True

        return cls.__instance

    def __init__(self, binder):
        if self.__init_flag:
            self.__init_flag = False
            super(BinderDispatcher, self).__init__()
            self >> event_hook(QtCore.QEvent.User, self.__bind_cls__)
        event = QtCore.QEvent(QtCore.QEvent.User)
        QtWidgets.QApplication.postEvent(self, event)

    def __bind_cls__(self):
        for module, data in self._trace_dict_.items():
            for cls_name, _data in data.items():
                if hasattr(module, cls_name):
                    cls = getattr(module, cls_name)
                    for _, binding in _data.items():
                        binding.cls = cls
        self._trace_dict_.clear()

    def dispatch(self, command, *args, **kwargs):
        method_dict = OrderedDict(inspect.getmembers(self, predicate=inspect.ismethod))
        method_dict.pop("dispatch")
        func = method_dict.get(command)
        if func is None:
            raise RuntimeError("Binder Action %s not found" % command)
        return func(*args, **kwargs)

    def dumper(self, db_name=None, filters=None):
        dumper = BinderDumper._dumper_dict_.get(id(self.binder))
        db_name = dumper.db_name if dumper else db_name
        if not db_name:
            uid = BinderCollector.get_last_key()
            binder_list = BinderCollector.Binders[uid]
            index = binder_list.index(self.binder)
            rd = random.Random()
            rd.seed(index)
            hex = uuid.UUID(int=rd.getrandbits(128)).hex
            path = inspect.stack()[-1][1]
            md5 = hashlib.md5("".join((path, hex)).encode("utf-8")).hexdigest()
            db_name = md5

        return BinderDumper(self.binder, db_name, filters)

    def fn_bind(self, attr=None):
        """fn_bind
        https://stackoverflow.com/a/13699329
        """

        def wrapper(func):
            binding = FnBinding(self.binder, func)
            function = func if six.callable(func) else func.__func__
            fn_name = attr if attr else function.__name__
            setattr(self.binder.__class__, fn_name, binding)

            stack = inspect.stack()[-2]
            cls_name = stack[3]
            frame = stack[0]
            module = inspect.getmodule(frame)

            self._trace_dict_[module][cls_name][fn_name] = binding

            return func

        return wrapper

    def dispatcher(self):
        return self


class BinderProxy(object):
    _var_dict_ = {}

    def __getitem__(self, key):
        return self._var_dict_.get(key)

    def __setitem__(self, key, value):
        var = self._var_dict_.get(key)
        var.set(value) if var else None

    def __setattr__(self, key, value):
        self._var_dict_[key] = value
        value = value if isinstance(value, Binding) else Binding(value)
        self.__dict__[key] = value


class BinderBase(object):
    _var_dict_ = {}
    _trace_flag_ = False
    _trace_setattr_ = []

    def __getitem__(self, key):
        val = self._var_dict_.get(key)
        if val is not None:
            return val
        else:
            return BindingProxy(self, key)

    def __setitem__(self, key, value):
        var = self._var_dict_.get(key)
        var.set(value) if var else None

    # def __getattr__(self, key):
    #     return BindingProxy(self, key)

    def __setattr__(self, key, value):
        if self._trace_flag_:
            self._trace_setattr_.append(key)
        binding = self._var_dict_.get(key)
        if binding:
            binding.set(value)
        else:
            # NOTE assign binding to class static member
            binding = Binding(value)
            binding.__binder__ = self
            self._var_dict_[key] = binding
            setattr(self.__class__, key, binding)
            if isinstance(value, FnBinding):
                value.connect_binder(key, self)

    def __call__(self, *args):
        # NOTE __call__ dispatch function avoid polluting local scope
        return BinderDispatcher(self).dispatch(*args)

    def __enter__(self):
        # NOTE support with statement for group indent
        self.proxy = BinderProxy()
        return self.proxy

    def __exit__(self, *args):
        # NOTE get the outer frame
        frame = inspect.currentframe().f_back

        for k, v in frame.f_locals.items():
            if self.proxy is v:
                for n, s in inspect.getmembers(self.proxy):
                    if isinstance(s, Binding):
                        self._var_dict_[n] = s
                        setattr(self.__class__, n, s)
                # NOTE overwrite the local scope like `return` action
                frame.f_locals.update({k: self})
                break


class Binder(BinderBase):
    def __new__(cls, *args, **kw):
        # NOTE spawn differenct class instance to contain static member
        class BinderInstance(BinderBase):
            _var_dict_ = {}

        instance = cls.__new__(BinderInstance)
        binder_list = BinderCollector().get_current_Binders()
        binder_list.append(instance)
        return instance


class GBinder(BinderBase):
    # NOTE Global Singleton
    __instance = None
    _var_dict_ = {}

    # TODO Group by arg string
    def __new__(cls, *args, **kw):
        if cls.__instance is None:
            cls.__instance = BinderBase.__new__(cls)
            binder_list = BinderCollector().get_current_Binders()
            binder_list.append(
                cls.__instance
            ) if cls.__instance not in binder_list else None
        return cls.__instance



# -*- coding: utf-8 -*-
"""

"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from types import BuiltinMethodType

__author__ = "timmyliang"
__email__ = "820472580@qq.com"
__date__ = "2020-11-02 23:47:53"

import os
import sys
import six
import json
import types
import inspect
from functools import partial

import Qt
from Qt import QtCore, QtWidgets, QtGui

from .binding import Binding
from .util import nestdict, defaultdict
from .hookconfig import CONFIG

HOOKS = nestdict()
_HOOKS_REL = nestdict()
qt_dict = {"QtWidgets.%s" % n: m for n, m in inspect.getmembers(QtWidgets)}
qt_dict.update({"QtCore.%s" % n: m for n, m in inspect.getmembers(QtCore)})
qt_dict.update({"QtGui.%s" % n: m for n, m in inspect.getmembers(QtGui)})


def byte2str(text):
    # NOTE compat python 2 and 3
    return str(text, encoding="utf-8") if sys.hexversion >= 0x3000000 else str(text)


def get_method_name(method):
    # NOTE compat Qt 4 and 5
    version = QtCore.qVersion()
    name = ""
    count = False
    if version.startswith("5"):
        name = method.name()
        count = method.parameterCount()
    elif version.startswith("4"):
        name = method.signature()
        name = name.split("(")[0]
        count = method.parameterNames()
    return byte2str(name), count

for name, member in qt_dict.items():
    # NOTE filter qt related func
    if not hasattr(member, "staticMetaObject"):
        continue
    meta_obj = getattr(member, "staticMetaObject")
    data = CONFIG.get(name, {})
    for method_name, method in inspect.getmembers(member, inspect.isroutine):
        if data.get(method_name):
            HOOKS[name][method_name] = {}
            if method_name.startswith("set"):
                _HOOKS_REL[method_name.lower()] = method_name

    # NOTE auto bind updater
    for i in range(meta_obj.propertyCount()):
        property = meta_obj.property(i)
        if not property.hasNotifySignal():
            continue
        property_name = property.name()
        method_name = _HOOKS_REL.get("set%s" % property_name.lower())
        data = HOOKS[name].get(method_name)
        if isinstance(data, dict):
            updater, _ = get_method_name(property.notifySignal())
            if updater:
                data.update({"updater": updater, "property": property_name})


class FuncHook(object):
    @classmethod
    def fix_cursor_position(cls, func, widget):
        """maintain the Qt edit cusorPosition after setting a new value"""

        def wrapper(*args, **kwargs):
            pos = widget.property("cursorPosition")
            res = func(*args, **kwargs)
            QtCore.QTimer.singleShot(
                0, lambda: widget.setProperty("cursorPosition", pos)
            ) if pos else None
            return res

        return wrapper

    @classmethod
    def auto_dump(cls, binding):
        """auto dump for two way binding"""
        
        from .constant import AUTO_DUMP

        binder = binding.__binder__
        if not AUTO_DUMP or not binder:
            return
        dumper = binder("dumper")
        for k, v in binder._var_dict_.items():
            if v is binding:
                dumper._filters_.add(k)
                break

    @classmethod
    def combine_args(cls, val, args):
        if isinstance(val, tuple):
            return val + args[1:]
        else:
            return (val,) + args[1:]
    
    def __init__(self, options):
        self.options = options

    def __call__(cls, func):
        @six.wraps(func)
        def wrapper(self,*args, **kwargs):
            if len(args) != 1:
                return func(self, *args, **kwargs)

            callback = args[0]

            if isinstance(callback, types.LambdaType):

                # NOTE get the running bindings (with __get__ method) add to Binding.TRACE_LIST
                with Binding.set_trace():
                    val = callback()

                def connect_callback(callback, args):
                    val = callback()
                    args = cls.combine_args(val, args)
                    func(self, *args, **kwargs)

                # NOTE register auto update
                _callback_ = partial(connect_callback, callback, args)
                for binding in Binding._trace_list_:
                    binding.connect(_callback_)

                args = cls.combine_args(val, args)

                # NOTE Single binding connect to the updater
                updater = cls.options.get("updater")
                prop = cls.options.get("property")
                getter = cls.options.get("getter")
                _getter_1 = getattr(self, getter) if getter else None
                _getter_2 = lambda: self.property(prop) if prop else None
                getter = _getter_1 if _getter_1 else _getter_2
                code = callback.__code__

                if (
                    updater
                    and getter
                    and len(Binding._trace_list_)
                    == 1  # NOTE only bind one response variable
                    and len(code.co_consts) == 1  # NOTE only bind directly variable
                ):
                    updater = getattr(self, updater)
                    updater.connect(
                        cls.fix_cursor_position(lambda *args: binding.set(getter()), self)
                    )
                    binding = Binding._trace_list_.pop()
                    QtCore.QTimer.singleShot(0, partial(cls.auto_dump, binding))

            return func(self, *args, **kwargs)

        return wrapper


def hook_initialize(hooks):
    """
    # NOTE Dynamic wrap the Qt Widget setter base on the HOOKS Definition
    """
    for widget, setters in hooks.items():
        lib, widget = widget.split(".")
        widget = getattr(getattr(Qt, lib), widget)
        for setter, options in setters.items():
            func = getattr(widget, setter)
            setattr(widget, setter, FuncHook(options)(func))
            


import os
import sys

# try:
#     import Qt
#     import six
# except:
DIR = os.path.dirname(__file__)
MODULE = os.path.join(DIR, "_vendor")
if MODULE not in sys.path:
    sys.path.insert(0,MODULE)
import Qt
import six


from .util import ListGet
from .hook import HOOKS, hook_initialize
from .binding import Model, Binding, FnBinding
from .binder import GBinder, Binder, BinderCollector
from .eventhook import QEventHook
from .decorator import inject


# NOTE hook Qt caller to accept lambda argument
hook_initialize(HOOKS)



# -*- coding: utf-8 -*-
"""

"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

__author__ = "timmyliang"
__email__ = "820472580@qq.com"
__date__ = "2020-11-15 20:11:41"

from functools import partial

import six
from .binding import Binding
from .util import ListGet
from Qt import QtWidgets


class HandlerBase(object):
    pass

class ItemMeta(type):
    
    def __getitem__(cls,item):
        cls.item = item
        return cls
class ItemConstructor(HandlerBase,six.with_metaclass(ItemMeta)):
    
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
    
    def __rrshift__(self, item_data):
        # binding = Binding._inst_.pop()
        
        # NOTE trace lamda function
        filters = self.kwargs.pop("__filters__", [i for i in range(len(item_data))])
        if callable(filters):
            # NOTE make sure handle list data
            filters = [item_data.index(d) for d in filters(item_data)]

        layout = self.kwargs.pop("__layout__", self.item.__layout__)
        binder_name = self.kwargs.pop("__binder__", self.item.__binder__)
        if not hasattr(layout,'__items__'):
            layout.__items__ = []
            
        for i, data in enumerate(item_data):
            widget = layout.__items__ >> ListGet(i)
            if not widget:
                widget = self.item(*self.args, **self.kwargs)
                layout.addWidget(widget)
                layout.__items__.append(widget)
                widget.__index__ = i

            widget.setVisible(i in filters)

            if hasattr(widget, binder_name):
                binder = getattr(widget, binder_name)
                for k, v in data.items():
                    val = getattr(binder, k)
                    setattr(binder, k, v)

        for i in range(len(item_data), len(layout.__items__)):
            widget = layout.__items__[i]
            widget.deleteLater()
        layout.__items__ = layout.__items__[: len(item_data)]


class GroupBoxBind(HandlerBase):
    def __init__(self, group):
        self.group = group

    def __rrshift__(self, binding):
        self.binding = Binding._inst_.pop()
        for rb in self.group.findChildren(QtWidgets.QRadioButton):
            rb.toggled.connect(partial(self.filter_state, rb))
        self.binding.connect(self.check_state) >> Call()

        return self.binding

    def filter_state(self, rb, state):
        if state:
            self.binding.set(rb.text().strip())

    def check_state(self):
        for rb in self.group.findChildren(QtWidgets.QRadioButton):
            if rb.text().strip() == self.binding.val:
                rb.setChecked(True)


class Call(HandlerBase):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __rrshift__(self, func):
        return func(*self.args, **self.kwargs)


class Set(HandlerBase):
    def __init__(self, val):
        self.val = val

    def __rrshift__(self, binding):
        binding = Binding._inst_.pop()
        binding.set(self.val)
        return binding


class Anim(HandlerBase):
    # TODO support anim value change
    def __init__(self, val):
        self.val = val

    def __rrshift__(self, binding):
        Binding._inst_
        return binding


# -*- coding: utf-8 -*-
"""

"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

__author__ = 'timmyliang'
__email__ = '820472580@qq.com'
__date__ = '2020-11-07 22:40:46'

from Qt import QtGui
from Qt import QtCore
from Qt import QtWidgets

from functools import partial

class CollapsibleWidget( QtWidgets.QWidget ):
    def __init__(self):
        super( CollapsibleWidget, self ).__init__()
        
    @staticmethod
    def install(btn,container,duration=300,expand_callback=None,collapse_callback=None):
        anim = QtCore.QPropertyAnimation(container, b"maximumHeight")
        
        anim.setDuration(duration)
        anim.setStartValue(0)
        anim.setEndValue(container.sizeHint().height())
        anim.finished.connect(lambda:container.setMaximumHeight(16777215) if not btn.toggle else None)

        btn.toggle = False
        btn.setText(u"▼ %s"%btn.text())
        def toggleFn(btn,anim):
            if btn.toggle:
                btn.toggle = False
                anim.setDirection(QtCore.QAbstractAnimation.Forward)

                anim.setEndValue(CollapsibleWidget.getHeightEndValue(container))
                anim.start()
                btn.setText(u"▼%s"%btn.text()[1:])
                btn.setStyleSheet('font:normal')
                if expand_callback:
                    expand_callback()
            else:
                btn.toggle = True
                anim.setDirection(QtCore.QAbstractAnimation.Backward)
                anim.setEndValue(container.sizeHint().height())
                anim.start()
                btn.setText(u"■%s"%btn.text()[1:])
                btn.setStyleSheet('font:bold')
                if collapse_callback:
                    collapse_callback()

        func = partial(toggleFn,btn,anim)
        btn.clicked.connect(func)
        return func

    @staticmethod
    def getHeightEndValue(widget):

        parent = widget.parent()
        total_height = parent.height()

        height = 0
        for child in parent.children():
            if child == widget or not hasattr(child,"height"):
                continue
            
            height += child.height()

        widget.updateGeometry()
        prefer = widget.sizeHint().height()
        height = total_height - height
        return height if height > prefer else prefer


# -*- coding: utf-8 -*-
"""

"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

__author__ = "timmyliang"
__email__ = "820472580@qq.com"
__date__ = "2020-11-15 20:56:50"

import six
import inspect
from .binder import Binder
from functools import wraps


def inject(binder):
    frame = inspect.currentframe().f_back
    for name, v in frame.f_locals.items():
        if v is binder:
            break

    def caller(func):
        @six.wraps(func)
        def wrapper(self, *args, **kwargs):
            # NOTE inject class binder to self binder
            # TODO FnBinding point to self
            _binder = Binder()
            for k, v in binder._var_dict_.items():
                setattr(_binder, k, v)
            setattr(self, name, _binder)
            res = func(self, *args, **kwargs)

            return res

        return wrapper

    return caller



# -*- coding: utf-8 -*-
"""
App Event Hook
hook the widget event rather than override class event method
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

__author__ = "timmyliang"
__email__ = "820472580@qq.com"
__date__ = "2020-11-02 23:47:53"

try:
    from collections import Iterable
except:
    from collections.abc import Iterable

import six
import inspect
from functools import partial

from Qt import QtCore
from Qt import QtWidgets
from Qt import QtGui


class QEventHook(QtCore.QObject):

    __invert_flag = False
    __instance = None
    __init_flag = False
    __hook = {}
    __invert_hook = {}
    __event = None
    __callbacks = None

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super(QEventHook, cls).__new__(cls)
            cls.__init_flag = True
        return cls.__instance

    def __init__(self):
        if self.__init_flag:
            self.__init_flag = False
            super(QEventHook, self).__init__()
            self.installEventFilter(self)
            event = QtCore.QEvent(QtCore.QEvent.User)
            QtCore.QCoreApplication.postEvent(self, event)

    def eventFilter(self, receiver, event):
        data = self.__hook.get(receiver)
        if data:
            callbacks = data.get(event.type(), [])
            for callback in callbacks:
                args = (receiver,)
                keywords = {}
                if isinstance(callback,partial):
                    args = callback.args
                    keywords = callback.keywords
                    callback  = callback.func
                    length = len(inspect.getargspec(callback).args)
                else:
                    length = len(inspect.getargspec(callback).args)
                count = length - 1 if inspect.ismethod(callback) else length
                args += (
                    event,
                )
                callback(*args[:count],**keywords)

        # NOTE invert event hook
        data = self.__invert_hook.get(receiver)
        if data is None:
            callbacks = self.__invert_hook.get(event.type(), [])
            for callback in callbacks:
                length = len(inspect.getargspec(callback).args)
                count = length - 1 if inspect.ismethod(callback) else length
                args = (
                    receiver,
                    event,
                )
                callback(*args[:count])

        if receiver is self and event.type() == QtCore.QEvent.User:
            self.removeEventFilter(self)
            app = QtWidgets.QApplication.instance()
            app.installEventFilter(self)
        return super(QEventHook, self).eventFilter(receiver, event)

    def __rrshift__(self, receiver):
        event = self.__event
        events = (
            (event,)
            if isinstance(event, six.string_types) or not isinstance(event, Iterable)
            else event
        )
        for event in events:
            event = getattr(QtCore.QEvent, event) if isinstance(event, str) else event
            if self.__invert_flag:
                self.add_invert_hook(receiver, event, self.__callbacks)
            else:
                self.add_hook(receiver, event, self.__callbacks)
        self.__invert_flag = False
        return receiver

    def __invert__(self):
        self.__invert_flag = True
        return self

    def __call__(self, event, callbacks):
        """__call__ drive add_hook

        https://doc.qt.io/qtforpython/PySide2/QtCore/QEvent.html

        :param event: QEvent Type
        :type event: str | QEvent.Type
        :param callbacks: callable list or callable object
        :type callbacks: callable
        """

        self.__event = event
        self.__callbacks = callbacks
        return self

    def add_hook(self, receiver, event, callbacks):
        """add_hook global hook

        https://doc.qt.io/qtforpython/PySide2/QtCore/QEvent.html

        :param receiver: object or widget
        :type receiver: QtCore.QObject
        :param event: QEvent Type , defaults to None
        :type event: str | QEvent.Type
        :param callbacks: callable list or callable object , defaults to None
        :type callbacks: callable
        """

        if not event or not callbacks:
            return
        self.__hook.setdefault(receiver, {})
        self.__hook[receiver].setdefault(event, [])
        self.__hook[receiver][event].extend(
            [c for c in callbacks if six.callable(c)]
            if isinstance(callbacks, list)
            else [callbacks]
        )

    def add_invert_hook(self, receiver, event, callbacks):
        if not event or not callbacks:
            return
        self.__invert_hook.setdefault(receiver, True)
        self.__invert_hook.setdefault(event, [])
        self.__invert_hook[event].extend(
            [c for c in callbacks if six.callable(c)]
            if isinstance(callbacks, list)
            else [callbacks]
        )

    def get_hook(self):
        return self.__hook

    def set_hook(self, hook):
        self.__hook.update(hook) 



# coding:utf-8
from __future__ import division, print_function

__author__ = "timmyliang"
__email__ = "820472580@qq.com"
__date__ = "2020-05-06 23:02:37"

"""

"""


import six
import sys
import time
import inspect
from functools import partial, wraps
from collections import OrderedDict
from contextlib import contextmanager
from Qt import QtCore, QtGui, QtWidgets
from .eventhook import QEventHook


event_hook = QEventHook()


class BindingBase(object):
    pass


class BindingProxy(BindingBase):
    def __init__(self, binder, attr):
        self.binder = binder
        self.attr = attr

    def __rrshift__(self, d):
        self.binder.__setattr__(self.attr, d)
        del self
        return d

    def __iter__(self):
        for attr in dir(self):
            yield attr


class FnBinding(BindingBase):
    
    def __init__(self, binder=None, func=None):
        from .binder import Binder
        self.cls = None
        self.binder = binder
        self.binded = isinstance(binder,Binder)
        if self.binded:
            self.func = func if six.callable(func) else func.__func__
            self.static = isinstance(self.func, staticmethod)
    
    def connect_binder(self, name, binder):
        """
        connect_binder automatically run by the BinderBase __setattr__
        """
        self.name = name
        self.binder = binder

    def __call__(self, *args, **kwargs):
        
        # NOTE if not initialize binded should bind func
        if not self.binded:
            self.binded = True
            func = args[0]
            self.func = func if six.callable(func) else func.__func__
            
            stack = inspect.stack()[-2]
            cls_name = stack[3]
            frame = stack[0]
            module = inspect.getmodule(frame)
            dispatcher = self.binder('dispatcher')
            dispatcher._trace_dict_[module][cls_name][self.func.__name__] = self

            return self.func
        
        arg = self.binder
        if self.cls:
            # TODO not very good solution | may be try __subclasshook__
            # NOTE Try to Get A Default Instance from binder
            for _, member in inspect.getmembers(self.binder, lambda f: not callable(f)):
                if type(member) is self.cls:
                    arg = member
                    return self.func(arg, *args, **kwargs)
        
        return self.func(*args, **kwargs)

    def __getitem__(self, attr):
        # NOTE special bind 
        attr = getattr(self.binder, attr) if type(attr) is str else attr

        @six.wraps(self.func)
        def wrapper(*args, **kw):
            if self.static:
                return self.func(*args, **kw)
            else:
                length = len(inspect.getargspec(self.func).args)
                return self.func(attr, *args[: length - 1], **kw)

        return wrapper


def notify(func):
    def wrapper(self, *args, **kwargs):
        res = func(self, *args, **kwargs)
        if hasattr(self, "STATE"):
            self.STATE.emitDataChanged()
            self.STATE.emit()
        return res

    return wrapper


class NotifyList(list):
    """
    https://stackoverflow.com/questions/13259179/list-callbacks
    """

    __repr__ = list.__repr__

    def __init__(self, val, STATE):
        super(NotifyList, self).__init__(val)
        self.STATE = STATE

    extend = notify(list.extend)
    append = notify(list.append)
    remove = notify(list.remove)
    pop = notify(list.pop)
    __iadd__ = notify(list.__iadd__)
    __imul__ = notify(list.__imul__)

    # Take care to return a new NotifyList if we slice it.
    if sys.version_info[0] < 3:
        __setslice__ = notify(list.__setslice__)
        __delslice__ = notify(list.__delslice__)

        def __getslice__(self, *args):
            return self.__class__(list.__getslice__(self, *args))

    __delitem__ = notify(list.__delitem__)

    def __getitem__(self, item):
        if isinstance(item, slice):
            return list(list.__getitem__(self, item))
        else:
            return list.__getitem__(self, item)

    @notify
    def __setitem__(self, key, value):
        if isinstance(value, dict):
            value = NotifyDict(value, self.STATE)
        elif isinstance(value, list):
            value = NotifyList(value, self.STATE)
        list.__setitem__(self, key, value)


class NotifyDict(OrderedDict):
    """
    https://stackoverflow.com/questions/5186520/python-property-change-listener-pattern
    """

    __repr__ = dict.__repr__

    def __init__(self, val, STATE):
        super(NotifyDict, self).__init__(val)
        self.STATE = STATE

    clear = notify(OrderedDict.clear)
    pop = notify(OrderedDict.pop)
    popitem = notify(OrderedDict.popitem)
    setdefault = notify(OrderedDict.setdefault)
    update = notify(OrderedDict.update)
    __delitem__ = notify(OrderedDict.__delitem__)

    @notify
    def __setitem__(self, key, value):
        if hasattr(self, "STATE"):
            if isinstance(value, dict):
                value = NotifyDict(value, self.STATE)
            elif isinstance(value, list):
                value = NotifyList(value, self.STATE)
        return OrderedDict.__setitem__(self, key, value)


class Binding(QtGui.QStandardItem, BindingBase):

    __trace = False
    __emit_flag = False
    _trace_list_ = []
    _inst_ = []

    __repr__ = lambda self: repr(self.val)
    __str__ = lambda self: str(self.val)

    operator_list = {
        "__add__": lambda self, x: self.val.__add__(x),
        "__sub__": lambda self, x: self.val.__sub__(x),
        "__mul__": lambda self, x: self.val.__mul__(x),
        "__floordiv__": lambda self, x: self.val.__floordiv__(x),
        "__truediv__": lambda self, x: self.val.__truediv__(x),
        "__mod__": lambda self, x: self.val.__mod__(x),
        "__pow__": lambda self, x: self.val.__pow__(x),
        "__lshift__": lambda self, x: self.val.__lshift__(x),
        "__rshift__": lambda self, x: self.val.__rshift__(x),
        "__and__": lambda self, x: self.val.__and__(x),
        "__xor__": lambda self, x: self.val.__xor__(x),
        "__or__": lambda self, x: self.val.__or__(x),
        "__iadd__": lambda self, x: self.val.__iadd__(x),
        "__isub__": lambda self, x: self.val.__isub__(x),
        "__imul__": lambda self, x: self.val.__imul__(x),
        "__idiv__": lambda self, x: self.val.__idiv__(x),
        "__ifloordiv__": lambda self, x: self.val.__ifloordiv__(x),
        "__imod__": lambda self, x: self.val.__imod__(x),
        "__ipow__": lambda self, x: self.val.__ipow__(x),
        "__ilshift__": lambda self, x: self.val.__ilshift__(x),
        "__irshift__": lambda self, x: self.val.__irshift__(x),
        "__iand__": lambda self, x: self.val.__iand__(x),
        "__ixor__": lambda self, x: self.val.__ixor__(x),
        "__ior__": lambda self, x: self.val.__ior__(x),
        "__neg__": lambda self, x: self.val.__neg__(x),
        "__pos__": lambda self, x: self.val.__pos__(x),
        "__abs__": lambda self, x: self.val.__abs__(x),
        "__invert__": lambda self, x: self.val.__invert__(x),
        "__complex__": lambda self, x: self.val.__complex__(x),
        "__int__": lambda self, x: self.val.__int__(x),
        "__long__": lambda self, x: self.val.__long__(x),
        "__float__": lambda self, x: self.val.__float__(x),
        "__oct__": lambda self, x: self.val.__oct__(x),
        "__hex__": lambda self, x: self.val.__hex__(x),
        "__lt__": lambda self, x: self.val.__lt__(x),
        "__le__": lambda self, x: self.val.__le__(x),
        "__eq__": lambda self, x: self.val.__eq__(x),
        "__ne__": lambda self, x: self.val.__ne__(x),
        "__ge__": lambda self, x: self.val.__ge__(x),
        "__gt__": lambda self, x: self.val.__gt__(x),
    }

    def __init__(self, val=None):
        super(Binding, self).__init__()
        self.val = self.retrieve2Notify(val)
        self.overrideOperator(self.val)
        self.event_loop = []
        self.__binder__ = None

    @classmethod
    @contextmanager
    def set_trace(cls):
        del cls._trace_list_[:]
        cls.__trace = True
        yield
        cls.__trace = False

    def __get__(self, instance, owner):
        self.__class__._inst_ = [self]
        self._trace_list_.append(self) if self.__trace else None
        return self.get()

    def __rrshift__(self, d):
        self.set(d)
        return d

    def set(self, value):
        # if value == self.get():
        #     return
        self.val = self.retrieve2Notify(value)
        self.overrideOperator(value)
        self.emitDataChanged()
        self.emit()

    def get(self):
        return self.val

    @classmethod
    def overrideOperator(cls, val):
        for attr, func in cls.operator_list.items():
            if attr in dir(val):
                setattr(cls, attr, func)

    def retrieve2Notify(self, val, initialize=True):
        """ convert to Notify type """
        itr = (
            six.iteritems(val)
            if isinstance(val, dict)
            else enumerate(val)
            if isinstance(val, list)
            else []
        )
        for k, v in itr:
            if isinstance(v, dict):
                self.retrieve2Notify(v, initialize=False)
                val[k] = NotifyDict(v, self)
            elif isinstance(v, list):
                self.retrieve2Notify(v, initialize=False)
                val[k] = NotifyList(v, self)

        if initialize:
            if isinstance(val, dict):
                return NotifyDict(val, self)
            elif isinstance(val, list):
                return NotifyList(val, self)
            else:
                return val

    def data(self, role):
        if role == QtCore.Qt.DisplayRole or role == QtCore.Qt.EditRole:
            return str(self.val)

    def setData(self, value, role=QtCore.Qt.EditRole):
        if role == QtCore.Qt.EditRole:
            self.set(value)
            return True
        return False

    def connect(self, callback):
        self.event_loop.append(callback)
        return callback

    def disconnect(self, callback):
        self.event_loop.remove(callback)

    def emit(self, *args, **kwargs):
        QtCore.QTimer.singleShot(0, lambda: self.run_event(*args, **kwargs))
        self.__emit_flag = True

    def run_event(self, *args, **kwargs):
        if self.__emit_flag:
            self.__emit_flag = False
            for callback in self.event_loop[:]:
                if six.callable(callback):
                    try:
                        callback(*args, **kwargs)
                    except:
                        self.event_loop.remove(callback)


class Model(QtCore.QAbstractItemModel):
    def __init__(self, source=None):
        super(Model, self).__init__()
        self.setSource(source)

    def dataChangedEmit(self):
        self.dataChanged.emit(QtCore.QModelIndex(), QtCore.QModelIndex())

    def setSource(self, source):
        # NOTE update source data remove old callback
        [
            item.disconnect(self.dataChangedEmit)
            for row in self._source
            for item in row
            if isinstance(item, Binding)
        ] if hasattr(self, "_source") else None

        self._source = (
            [
                item
                if isinstance(item, list)
                else [item if isinstance(item, Binding) else item]
                for item in source
            ]
            if source
            else []
        )

        # NOTE add data update callback
        for row in self._source:
            for item in row:
                if isinstance(item, Binding):
                    item.connect(self.dataChangedEmit)

        # NOTE fill None to the empty cell
        columnCount = max([len(row) for row in self._source])
        for row in self._source:
            rowCount = len(row)
            if rowCount < columnCount:
                row.extend([None for i in range(columnCount - rowCount)])

    def index(self, row, column, parent=QtCore.QModelIndex()):
        return self.createIndex(row, column)

    def parent(self, index):
        return QtCore.QModelIndex()

    def rowCount(self, parent):
        return len(self._source) if parent.row() == -1 else 0

    def columnCount(self, parent):
        return max([len(row) for row in self._source])

    def item(self, row, column):
        row_list = next(iter(self._source[row:]), None)
        if row_list is None:
            return
        return next(iter(row_list[column:]), None)

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if role == QtCore.Qt.DisplayRole or role == QtCore.Qt.EditRole:
            if index.isValid():
                item = self.item(index.row(), index.column())
                return (
                    item.val
                    if isinstance(item, Binding)
                    else item.text()
                    if isinstance(item, QtGui.QStandardItem)
                    else str(item)
                )

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        if role == QtCore.Qt.EditRole:
            if index.isValid():
                row = index.row()
                column = index.column()
                item = self.item(row, column)
                if isinstance(item, Binding):
                    item.set(value)
                elif isinstance(item, QtGui.QStandardItem):
                    item.setText(value)
                else:
                    self._source[row][column] = value
                self.dataChanged.emit(index, index)
                return True
        return False

    def flags(self, index):
        return (
            QtCore.Qt.ItemIsEditable
            | QtCore.Qt.ItemIsEnabled
            | QtCore.Qt.ItemIsSelectable
        )

    def get(self):
        return self._source



# -*- coding: utf-8 -*-
"""

"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

__author__ = 'timmyliang'
__email__ = '820472580@qq.com'
__date__ = '2020-11-28 10:38:36'


import os
import sys
import requests
from setuptools import setup, find_packages
from QBinder.constant import VERSION
if sys.version_info[0] < 3:
    from codecs import open

# NOTE 将markdown格式转换为rst格式
def md_to_rst(from_file, to_file):
    r = requests.post(url='http://c.docverter.com/convert',
                      data={'to':'rst','from':'markdown'},
                      files={'input_files[]':open(from_file,'rb')})
    if r.ok:
        with open(to_file, "wb") as f:
            f.write(r.content)


md_to_rst("README.md", "README.rst")


if os.path.exists('README.rst'):
    long_description = open('README.rst', encoding="utf-8").read()
else:
	long_description = 'Global Data Binding for Python Qt framework'

long_description = "Global Data Binding for Python Qt framework"

setup(
    name="QBinder",
    version=VERSION,
    keywords=("pip", "PyQt", "PySide", "Qt", "DataBinding", "Binding", "Binder"),
    description="Global Data Binding for Python Qt framework",
    long_description=long_description,
    license="MIT Licence",
    url="https://github.com/FXTD-ODYSSEY/QBinder",
    author="timmyliang",
    author_email="820472580@qq.com",
    packages=find_packages(),
    include_package_data=True,
    platforms="any",
    install_requires=[
        "six",
        "Qt.py",
    ],
)


# -*- coding: utf-8 -*-
"""

"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

__author__ = "timmyliang"
__email__ = "820472580@qq.com"
__date__ = "2020-11-07 22:26:11"

VERSION = "1.0.1"
AUTO_DUMP = True



# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

__author__ = 'timmyliang'
__email__ = '820472580@qq.com'
__date__ = '2020-11-05 09:07:49'


class ListGet(object):
    """ListGet 
    safe get list index like dict.get
    https://stackoverflow.com/a/26849807
    
    >>> a = [1,2,3]
    >>> val = a >> ListGet(2)
    >>> print(val)
    3
    """
    
    def __init__(self,index, defualt=None):
        self.index = index
        self.defualt = defualt
    
    def __rrshift__(self, d):
        d = d if isinstance(d,list) else [d]
        return next(iter(d[self.index:]), self.defualt)
    
if __name__ == "__main__":
    import doctest
    doctest.testmod()